/******************************************************************************
 * Reference:                                                                 *
 * https://eli.thegreenplace.net/2013/09/16/analyzing-function-cfgs-with-llvm *
 ******************************************************************************/

#include <system_error>
#include <string>
#include <set>
#include <utility>
#include <map>

#include "llvm/Pass.h"
#include "llvm/IR/Function.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/IR/LegacyPassManager.h"
#include "llvm/Transforms/IPO/PassManagerBuilder.h"
#include "llvm/IR/Function.h"
#include "llvm/ADT/StringRef.h"
#include "llvm/Support/FileSystem.h"
#include "llvm/Transforms/Utils/BasicBlockUtils.h"

using namespace llvm;
using namespace llvm::sys::fs;

namespace {
  class Graph : public FunctionPass {
    private:
      /* Keep track of visited blocks during DFS. */
      std::set<BasicBlock *> visitedBlocks;

      /* Set of directed edges (a->b). */
      std::vector<std::pair<BasicBlock *, BasicBlock *> > edges;

      /* Each basic block is associated with an ID. */
      std::map<BasicBlock *, unsigned int> nodeId;

      void dfs(BasicBlock *BB) {
        /* Skip the node if it was already reached. */
        if (!std::get<1>(visitedBlocks.insert(BB))) {
          return;
        }

        /* Assign an ID number. */
        nodeId[BB] = visitedBlocks.size();
        
        /*
         * Get the terminating instruction of the block.
         * The successors are all the blocks which may be entered from the current basic block.
         */
        const TerminatorInst *terminInst = BB->getTerminator();
        for (unsigned index = 0,
            nSucc = terminInst->getNumSuccessors();
            index < nSucc; ++index) {
          BasicBlock *succBlock = terminInst->getSuccessor(index);
          edges.push_back(std::make_pair(BB, succBlock));
          dfs(succBlock);
        }
      }

      /*
       * Dump the representation of the CFG in cfg/functionName.cfg
       *
       * The first line of the file contains 'n', the number of nodes.
       * Each of the following lines contains two numbers separated
       * by a space, 'a b', 1<=a,b<=n, denoting the directed edge
       * a->b.
       */
      void dumpCFG(std::string functionName) {
        functionName.insert(0, "cfg/");
        functionName.append(".cfg");
        std::error_code e;
        raw_fd_ostream out(StringRef(functionName), e, OpenFlags::F_RW);
        out << visitedBlocks.size() << '\n';
        for (const auto &it : edges) {
          out << nodeId[it.first] << ' ' << nodeId[it.second] << '\n';
        }
        out.close();
      }

      void clear() {
        visitedBlocks.clear();
        edges.clear();
        nodeId.clear();
      }

    public:
      static char ID;
      Graph() : FunctionPass(ID) {}

      bool runOnFunction(Function &F) override {
        /* Build the CFG by running DFS from the entry block. */
        BasicBlock *bb = &F.getEntryBlock();
        dfs(bb);
        dumpCFG(F.getName().str());
        clear();
        return false;
      }
  };
}

char Graph::ID = 0;

static void registerGraphPass(const PassManagerBuilder &,
    legacy::PassManagerBase &PM) {
  PM.add(new Graph());
}

static RegisterStandardPasses
    RegisterMyPass0(PassManagerBuilder::EP_EnabledOnOptLevel0, registerGraphPass);
